import jax.numpy as np
import jax
import variationaldist as vd
from jax.flatten_util import ravel_pytree
import functools
import ais_utils


def initialize(dim_g, dim_l, num_l, model_params, vdmode_g=1, vdmode_l=1, stl_g=False, stl_l=False, vdparams_g=None,
	vdparams_l=None, nbridges=0, lfsteps=1, eps=0.01, eta=0.5, mode_eps='nodep_vec', mdparams=None, ngridb=32,
	mgridref_y=None, trainable = ['vd_g', 'vd_l', 'eps', 'eta']):
	
	params_train = {}
	params_notrain = {}

	if 'vd_g' in trainable:
		params_train['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_train['vd_g'] = vd.initialize_global(dim_g, vdmode_g)
	else:
		params_notrain['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_notrain['vd_g'] = vd.initialize_global(dim_g, vdmode_g)

	if 'vd_l' in trainable:
		params_train['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_train['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)
	else:
		params_notrain['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_notrain['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)

	if mode_eps == "nodep_vec":
		if 'eps' in trainable:
			params_train['eps'] = eps * np.ones((num_l, dim_l))
		else:
			params_notrain['eps'] = eps * np.ones((num_l, dim_l))
	elif mode_eps == "nodep":
		if 'eps' in trainable:
			params_train['eps'] = eps
		else:
			params_notrain['eps'] = eps
	else:
		raise NotImplementedError("Mode eps not implemented.")

	if 'model_params' in trainable:
		params_train['model'] = model_params
	else:
		params_notrain['model'] = model_params

	# TODO: Could do also vector for this?
	if 'eta' in trainable:
		params_train['eta'] = eta
	else:
		params_notrain['eta'] = eta

	# Everything related to betas
	if mgridref_y is not None:
		ngridb = mgridref_y.shape[0] - 1
	else:
		if nbridges < ngridb:
			ngridb = nbridges
		mgridref_y = np.ones(ngridb + 1) * 1.
	params_notrain['gridref_x'] = np.linspace(0, 1, ngridb + 2)
	params_notrain['target_x'] = np.linspace(0, 1, nbridges + 2)[1:-1]
	if 'mgridref_y' in trainable:
		params_train['mgridref_y'] = mgridref_y
	else:
		params_notrain['mgridref_y'] = mgridref_y

	params_fixed = (dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, nbridges, lfsteps, mode_eps)
	params_flat, unflatten = ravel_pytree((params_train, params_notrain))
	return params_flat, unflatten, params_fixed





def compute_ratio(seed, params_flat, unflatten, params_fixed, log_prob_model, indices):
	params_train, params_notrain = unflatten(params_flat)
	params_notrain = jax.lax.stop_gradient(params_notrain)
	params = {**params_train, **params_notrain}
	dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, nbridges, lfsteps, mode_eps = params_fixed

	if nbridges >= 1:
		gridref_y = np.cumsum(params['mgridref_y']) / np.sum(params['mgridref_y'])
		gridref_y = np.concatenate([np.array([0.]), gridref_y])
		betas = np.interp(params['target_x'], params['gridref_x'], gridref_y)

	rng_key_gen = jax.random.PRNGKey(seed)

	# Sample z_g and initialize weight
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	z_g = vd.sample_rep_global(rng_key, vdmode_g, params['vd_g'])
	params_g = params['vd_g'] if not stl_g else jax.lax.stop_gradient(params['vd_g'])
	w = -vd.log_prob_global(vdmode_g, params_g, z_g)

	# Sample z_l and update weight
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	z_l_sub = vd.sample_rep_local(rng_key_gen, vdmode_l, params['vd_l'], indices)
	params_l = params['vd_l'] if not stl_l else jax.lax.stop_gradient(params['vd_l'])
	w = w - vd.log_prob_local(vdmode_l, params_l, z_l_sub, indices)

	# Evolve UHA and update weight
	if nbridges >= 1:
		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		z_l_sub, w_mom = ais_utils.evolve(z_g, z_l_sub, indices, betas, params, rng_key, params_fixed, log_prob_model,
			num_l * 1. / indices.shape[0])
		w = w + w_mom

	# Evaluate model and update weight
	w = w + log_prob_model(z_g, z_l_sub, params["model"], indices)
	return w


def compute_avg_ratio(seeds, params_flat, unflatten, params_fixed, log_prob_model, indices):
	ws = jax.vmap(compute_ratio, in_axes = (0, None, None, None, None, None))(seeds, params_flat, unflatten, params_fixed,
		log_prob_model, indices)
	return -ws.mean()

