import jax.numpy as np
import jax
import variationaldist as vd
from jax.flatten_util import ravel_pytree
import functools
import ais_utils
import jax.scipy as sp


def initialize(dim_g, dim_l, num_l, model_params, vdmode_g=1, vdmode_l=1, stl_g=False, stl_l=False, vdparams_g=None,
	vdparams_l=None, K=0, trainable = ['vd_g', 'vd_l']):
	
	params_train = {}
	params_notrain = {}

	if 'vd_g' in trainable:
		params_train['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_train['vd_g'] = vd.initialize_global(dim_g, vdmode_g)
	else:
		params_notrain['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_notrain['vd_g'] = vd.initialize_global(dim_g, vdmode_g)

	if 'vd_l' in trainable:
		params_train['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_train['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)
	else:
		params_notrain['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_notrain['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)

	if 'model_params' in trainable:
		params_train['model'] = model_params
	else:
		params_notrain['model'] = model_params

	params_fixed = (dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, K, 0., "none")
	params_flat, unflatten = ravel_pytree((params_train, params_notrain))
	return params_flat, unflatten, params_fixed




def compute_log_weight(seed, params_flat, unflatten, params_fixed, log_prob_model, indices):
	params_train, params_notrain = unflatten(params_flat)
	params_notrain = jax.lax.stop_gradient(params_notrain)
	params = {**params_train, **params_notrain}
	dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, K, _, _ = params_fixed

	rng_key_gen = jax.random.PRNGKey(seed)

	# Sample z_g and initialize weight
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	z_g = vd.sample_rep_global(rng_key, vdmode_g, params['vd_g'])
	params_g = params['vd_g'] if not stl_g else jax.lax.stop_gradient(params['vd_g'])
	w = -vd.log_prob_global(vdmode_g, params_g, z_g)

	# Sample z_l and update weight
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	z_l_sub = vd.sample_rep_local(rng_key_gen, vdmode_l, params['vd_l'], indices)
	params_l = params['vd_l'] if not stl_l else jax.lax.stop_gradient(params['vd_l'])
	w = w - vd.log_prob_local(vdmode_l, params_l, z_l_sub, indices)

	# Evaluate model and update weight
	w = w + log_prob_model(z_g, z_l_sub, params["model"], indices)
	return w



def compute_ratio(seed, params_flat, unflatten, params_fixed, log_prob_model, indices):
	params_train, params_notrain = unflatten(params_flat)
	params_notrain = jax.lax.stop_gradient(params_notrain)
	params = {**params_train, **params_notrain}
	dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, K, _, _ = params_fixed

	rng_key_gen = jax.random.PRNGKey(seed)
	seeds = jax.random.randint(rng_key_gen, (K,), 1, 1e6)

	log_weights = jax.vmap(compute_log_weight, in_axes = (0, None, None, None, None, None))(seeds, params_flat, unflatten,
		params_fixed, log_prob_model, indices)
	K = log_weights.shape[0]

	return np.log(1./K) + sp.special.logsumexp(log_weights)



def compute_avg_ratio(seeds, params_flat, unflatten, params_fixed, log_prob_model, indices):
	ws = jax.vmap(compute_ratio, in_axes = (0, None, None, None, None, None))(seeds, params_flat, unflatten, params_fixed,
		log_prob_model, indices)
	return -ws.mean()

