import jax.numpy as np
import jax
import variationaldist as vd
from jax.flatten_util import ravel_pytree
import functools
import ais_utils
import jax.scipy as sp


def initialize(dim_g, dim_l, num_l, model_params, vdmode_g=1, vdmode_l=1, stl_g=False, stl_l=False, vdparams_g=None,
	vdparams_l=None, K=0, trainable = ['vd_g', 'vd_l']):
	
	params_train = {}
	params_notrain = {}

	if 'vd_g' in trainable:
		params_train['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_train['vd_g'] = vd.initialize_global(dim_g, vdmode_g)
	else:
		params_notrain['vd_g'] = vdparams_g
		if vdparams_g is None:
			params_notrain['vd_g'] = vd.initialize_global(dim_g, vdmode_g)

	if 'vd_l' in trainable:
		params_train['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_train['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)
	else:
		params_notrain['vd_l'] = vdparams_l
		if vdparams_l is None:
			params_notrain['vd_l'] = vd.initialize_local(dim_l, num_l, vdmode_l)

	if 'model_params' in trainable:
		params_train['model'] = model_params
	else:
		params_notrain['model'] = model_params

	params_fixed = (dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, K, 0., "none")
	params_flat, unflatten = ravel_pytree((params_train, params_notrain))
	return params_flat, unflatten, params_fixed





def compute_ratio(seed, params_flat, unflatten, params_fixed, log_prob_model_global, log_prob_model_local_ind, indices):
	def sample_local_aux(seed, vdmode_l, params_vdl, indices):
		rng_key_gen = jax.random.PRNGKey(seed)
		z_l_sub = vd.sample_rep_local(rng_key_gen, vdmode_l, params_vdl, indices)
		return z_l_sub

	def local_augmented_bound(z_g, z_l_K, params_vdl, i):
		# z_l_K has shape (K, dim_l), all samples from group i
		def get_log_weight(z_l):
			# z_l has shape (dim_l), all samples from group i
			w = log_prob_model_local_ind(z_g, z_l, params["model"], i)
			w = w - vd.log_prob_local_ind(vdmode_l, params_vdl, z_l, i)
			return w

		log_weights = jax.vmap(get_log_weight, in_axes = (0,))(z_l_K)
		K = log_weights.shape[0]
		return np.log(1./K) + sp.special.logsumexp(log_weights)


	params_train, params_notrain = unflatten(params_flat)
	params_notrain = jax.lax.stop_gradient(params_notrain)
	params = {**params_train, **params_notrain}
	dim_g, dim_l, num_l, vdmode_g, vdmode_l, stl_g, stl_l, K, _, _ = params_fixed

	rng_key_gen = jax.random.PRNGKey(seed)

	# Sample z_g and initialize weight
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	z_g = vd.sample_rep_global(rng_key, vdmode_g, params['vd_g'])
	params_g = params['vd_g'] if not stl_g else jax.lax.stop_gradient(params['vd_g'])
	w = -vd.log_prob_global(vdmode_g, params_g, z_g)

	# Eval global part of model at z_g
	w += log_prob_model_global(z_g, params["model"])

	# Sample z_l K times
	rng_key, rng_key_gen = jax.random.split(rng_key_gen)
	seeds = jax.random.randint(rng_key, (K,), 1, 1e6)
	z_l_sub_K = jax.vmap(sample_local_aux, in_axes = (0, None, None, None))(seeds, vdmode_l, params['vd_l'], indices)
	# Shape (K, num_local_subs, dim_l)

	# Compute, for each i, the local augmented bound
	local_bounds = jax.vmap(local_augmented_bound, in_axes = (None, 1, None, 0))(z_g, z_l_sub_K, params['vd_l'], indices)
	# Has shape indices.shape[0]

	f = num_l * 1. / indices.shape[0]

	return w + local_bounds.sum() * f


def compute_avg_ratio(seeds, params_flat, unflatten, params_fixed, log_prob_model_global, log_prob_model_local_ind, indices):
	ws = jax.vmap(compute_ratio, in_axes = (0, None, None, None, None, None, None))(seeds, params_flat, unflatten, params_fixed,
		log_prob_model_global, log_prob_model_local_ind, indices)
	return -ws.mean()

