import jax.numpy as np
import numpyro.distributions as npdist
import numpy as onp
import jax
import pickle
from .reparam_aux import apply_transform


def sample(savedir, num_groups, dim_z, obs_per_group, mode, reps=1, seed=1):
	assert num_groups == len(obs_per_group)
	onp.random.seed(seed)
	rng_key_gen = jax.random.PRNGKey(seed)

	max_Ni = max(obs_per_group)

	# Build mask
	mask = np.array([[1.] * int(n) + [0.] * int((max_Ni - n)) for n in obs_per_group])

	info = (num_groups, dim_z, obs_per_group, reps, seed, mask)
	name = "g%i_z%i_%s" % (num_groups, dim_z, mode)

	samples = {}

	for i in range(reps):
		# Build side info
		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		x = jax.random.normal(rng_key, shape = (num_groups, max_Ni, dim_z))
		x = x * mask[:, :, None] # Zero out non-observations for groups

		# Sample
		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		mean_z = npdist.Normal(loc = 0., scale = 1.).sample(rng_key)
		
		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		logscale_z = npdist.Normal(loc = 0., scale = 1.).sample(rng_key)

		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		logscale_y = npdist.Normal(loc = 0., scale = 1.).sample(rng_key)

		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		z = npdist.Normal(loc = mean_z * np.ones((num_groups, dim_z)), scale = np.exp(logscale_z)).sample(rng_key)

		mean_y = jax.vmap(np.dot, in_axes=(0, 0))(x, z)

		rng_key, rng_key_gen = jax.random.split(rng_key_gen)
		y = npdist.Normal(loc = mean_y, scale = np.exp(logscale_y)).sample(rng_key)
		y = y * mask

		samples[i] = {}
		samples[i]["x"] = x
		samples[i]["y"] = y
		samples[i]["mean_z"] = mean_z
		samples[i]["logscale_z"] = logscale_z
		samples[i]["logscale_y"] = logscale_y
		samples[i]["z"] = z

	full_name = "synthetic_%s.pkl" % name
	with open("%s%s" % (savedir, full_name), "wb") as f:
		print(full_name)
		pickle.dump((info, samples), f)



def load_model(name, index, param):
	def log_prob_local_ind(z_g, z_l, params, i):
		# z_l is a vector (dim_z)
		x_l = x[i, :, :] # Matrix (max(N_i), d_z)
		y_l = y[i, :] # Vector (max(N_i))
		mask_l = mask[i, :] # Vector (max(N_i)), with N_i ones followed by zeros
		N_i = mask_l.sum() # Number of observations in this group

		mean_z = z_g[0]
		logscale_z = z_g[1]
		logscale_y = z_g[2]

		mean_obs = np.dot(x_l, z_l) # Vector (N_i)
		scale_obs = np.exp(logscale_y) # Scalar

		lp_z = npdist.Normal(loc = mean_z, scale = np.exp(logscale_z)).log_prob(z_l)
		lp_obs = npdist.Normal(loc = mean_obs, scale = scale_obs).log_prob(y_l) * mask_l
		return lp_z.sum() + lp_obs.sum()

	def log_prob_global(z_g, params):
		mean_z = z_g[0]
		logscale_z = z_g[1]
		logscale_y = z_g[2]

		lp = npdist.Normal(loc = 0., scale = 1.).log_prob(mean_z).sum()
		lp = lp + npdist.Normal(loc = 0., scale = 1.).log_prob(logscale_z).sum()
		lp = lp + npdist.Normal(loc = 0., scale = 1.).log_prob(logscale_y).sum()
		return lp

	def log_prob(z_g, z_l_sub, params, indices, factor=True):
		lp = log_prob_global(z_g, params)

		f = jax.lax.cond(factor, lambda : num_l * 1. / indices.shape[0], lambda : 1.)

		lp_local = jax.vmap(log_prob_local_ind, in_axes = (None, 0, None, 0))(z_g, z_l_sub, params, indices)
		lp = lp + lp_local.sum() * f
		return lp



	def reparam(z_g, z_l_sub, params, indices, factor=True):
		log_det_jac = 0.
		params_g = params["z_g"]
		params_l_sub = params["z_l"][indices, :]

		z_g_out, log_det_jac_g = apply_transform(z_g, np.zeros(3), np.ones(3), params_g)
		log_det_jac += log_det_jac_g

		f = jax.lax.cond(factor, lambda : num_l * 1. / indices.shape[0], lambda : 1.)

		mean_z_aux = z_g_out[0] * np.ones((indices.shape[0], dim_l))
		scale_z_aux = np.exp(z_g_out[1])
		z_l_sub_out, log_det_jac_l_sub = apply_transform(z_l_sub, mean_z_aux, scale_z_aux, params_l_sub)
		log_det_jac += log_det_jac_l_sub * f

		return z_g_out, z_l_sub_out, log_det_jac

	def reparam_local(z_g, z_l, params, i):
		log_det_jac = 0.
		params_l_sub = params["z_l"][i, :]

		z_g_out, _ = reparam_global(z_g, params)

		mean_z_aux = z_g_out[0] * np.ones(dim_l)
		scale_z_aux = np.exp(z_g_out[1])
		z_l_out, log_det_jac_l = apply_transform(z_l, mean_z_aux, scale_z_aux, params_l_sub)
		log_det_jac += log_det_jac_l

		return z_g_out, z_l_out, log_det_jac

	def reparam_global(z_g, params):
		log_det_jac = 0.
		params_g = params["z_g"]

		z_g_out, log_det_jac_g = apply_transform(z_g, np.zeros(3), np.ones(3), params_g)
		log_det_jac += log_det_jac_g

		return z_g_out, log_det_jac

	def log_prob_final(z_g, z_l_sub, params, indices, factor=True):
		z_g_aux, z_l_sub_aux, log_det_jac = reparam(z_g, z_l_sub, params, indices, factor)
		return log_prob(z_g_aux, z_l_sub_aux, params, indices, factor) + log_det_jac

	def log_prob_local_ind_final(z_g, z_l, params, i):
		z_g_aux, z_l_aux, log_det_jac = reparam_local(z_g, z_l, params, i)
		return log_prob_local_ind(z_g_aux, z_l_aux, params, i) + log_det_jac

	def log_prob_global_final(z_g, params):
		z_g_aux, log_det_jac = reparam_global(z_g, params)
		return log_prob_global(z_g_aux, params) + log_det_jac

	with open("models/datasets/%s.pkl" % name, "rb") as f:
		info, samples = pickle.load(f)

	num_groups, dim_z, obs_per_group, _, _, mask = info	
	mask = np.array(mask)

	dim_g = 3

	dim_l = dim_z
	num_l = num_groups

	factor_init = 0. if param == 0 else 1.
	params = {}
	params["z_g"] = np.ones(3) * factor_init
	params["z_l"] = np.ones((num_l, dim_l)) * factor_init

	x = np.array(samples[index]["x"])
	y = np.array(samples[index]["y"])

	return log_prob_final, log_prob_global_final, log_prob_local_ind_final, dim_g, dim_l, num_l, params



